@@ -101,6 +101,7 @@ oracleSession
 	OCISvcCtx *svchp = NULL;
 	OCIServer *srvhp = NULL;
 	OCISession *userhp = NULL;
+	OCITrans *txnhp = NULL;
 	oracleSession *session;
 	struct envEntry *envp;
 	struct srvEntry *srvp;
@@ -317,6 +318,29 @@ oracleSession
 				oraMessage);
 		}
 
+		/* create transaction handle */
+		if (checkerr(
+			OCIHandleAlloc((dvoid *) envhp, (dvoid **) &txnhp,
+				(ub4) OCI_HTYPE_TRANS,
+				(size_t) 0, (dvoid **) 0),
+			(dvoid *)envhp, OCI_HTYPE_ENV) != OCI_SUCCESS)
+		{
+			oracleError(FDW_UNABLE_TO_ESTABLISH_CONNECTION,
+				"error connecting to Oracle: OCIHandleAlloc failed to allocate transaction handle",
+				oraMessage);
+		}
+
+		/* set transaction handle in service handle */
+		if (checkerr(
+			OCIAttrSet(svchp, OCI_HTYPE_SVCCTX, txnhp, 0,
+				OCI_ATTR_TRANS, errhp),
+			(dvoid *)errhp, OCI_HTYPE_ERROR) != OCI_SUCCESS)
+		{
+			oracleError(FDW_UNABLE_TO_ESTABLISH_CONNECTION,
+				"error connecting to Oracle: OCIAttrSet failed to set transaction handle in service handle",
+				oraMessage);
+		}
+
 		/* create session handle */
 		if (checkerr(
 			OCIHandleAlloc((dvoid *) envhp, (dvoid **) &userhp,
@@ -417,6 +441,16 @@ oracleSession
 		srvp->connlist = connp;
 	}
 
+	/* start a "serializable" (= repeatable read) transaction */
+    if (checkerr(
+        OCITransStart(svchp, errhp, (uword)0, OCI_TRANS_SERIALIZABLE),
+        (dvoid *)errhp, OCI_HTYPE_ERROR) != OCI_SUCCESS)
+    {
+        oracleError(FDW_UNABLE_TO_ESTABLISH_CONNECTION,
+            "error connecting to Oracle: OCITransStart failed to start a transaction",
+            oraMessage);
+    }
+
 	/* palloc a data structure containing server and session handle */
 	session = oracleAlloc(sizeof(struct oracleSession));
 	session->envhp = envhp;
@@ -440,6 +474,16 @@ oracleReleaseSession(oracleSession *session, struct oraTable *oraTable, int clos
 	/* close the statement, if any */
 	oracleCloseStatement(session, oraTable);
 
+	/* commit the current transaction */
+    if (checkerr(
+        OCITransCommit(session->svchp, session->errhp, OCI_DEFAULT),
+        (dvoid *)session->errhp, OCI_HTYPE_ERROR) != OCI_SUCCESS)
+    {
+        oracleError(FDW_UNABLE_TO_CREATE_EXECUTION,
+            "error committing transaction: OCITransCommit failed",
+            oraMessage);
+    }
+
 	/* close the session if requested */
 	if (close)
 		closeSession(session->envhp, session->srvhp, session->userhp, 1);
@@ -472,6 +516,7 @@ oracleCloseStatement(oracleSession *session, const struct oraTable *oraTable)
 		{
 			if (oraTable->cols[i]->used
 				&& (oraTable->cols[i]->oratype == ORA_TYPE_BLOB
+					|| oraTable->cols[i]->oratype == ORA_TYPE_BFILE
 					|| oraTable->cols[i]->oratype == ORA_TYPE_CLOB))
 			{
 				(void)OCIDescriptorFree((dvoid *)oraTable->cols[i]->val, OCI_DTYPE_LOB);
@@ -869,12 +914,17 @@ struct oraTable
 				reply->cols[i-1]->val_size = charsize * 4 + 1;
 				break;
 			case SQLT_BLOB:
-			case SQLT_BFILE:
-				/* BLOB and BFILE */
+				/* BLOB */
 				reply->cols[i-1]->oratype = ORA_TYPE_BLOB;
 				/* for LOB columns, "val" will contain a pointer to the locator */
 				reply->cols[i-1]->val_size = sizeof(OCILobLocator *);
 				break;
+			case SQLT_BFILE:
+				/* BFILE */
+				reply->cols[i-1]->oratype = ORA_TYPE_BFILE;
+				/* for LOB columns, "val" will contain a pointer to the locator */
+				reply->cols[i-1]->val_size = sizeof(OCILobLocator *);
+				break;
 			case SQLT_CLOB:
 			case SQLT_CFILE:
 				/* CLOB and CFILE */
@@ -1589,40 +1639,16 @@ oracleExecuteQuery(oracleSession *session, const char *query, const struct oraTa
 			switch (oraTable->cols[i]->oratype)
 			{
 				case ORA_TYPE_BLOB:
-				case ORA_TYPE_BFILE:
 					type = SQLT_BLOB;
-
-					/* "val" will contain the pointer to the LOB locator */
-					if (checkerr(
-						OCIDescriptorAlloc((const dvoid *)session->envhp,
-							(dvoid **)oraTable->cols[i]->val,
-							OCI_DTYPE_LOB, 0, NULL),
-						(dvoid *)session->envhp, OCI_HTYPE_ENV) != OCI_SUCCESS)
-					{
-						oracleCloseStatement(session, NULL);
-						oracleError(FDW_UNABLE_TO_CREATE_EXECUTION,
-							"error executing query: OCIDescriptorAlloc failed to allocate LOB descriptor",
-							oraMessage);
-					}
+					break;
+				case ORA_TYPE_BFILE:
+					type = SQLT_BFILE;
 					break;
 				case ORA_TYPE_CLOB:
 					type = SQLT_CLOB;
-
-					/* "val" will contain the pointer to the LOB locator */
-					if (checkerr(
-						OCIDescriptorAlloc((const dvoid *)session->envhp,
-							(dvoid **)oraTable->cols[i]->val,
-							OCI_DTYPE_LOB, 0, NULL),
-						(dvoid *)session->envhp, OCI_HTYPE_ENV) != OCI_SUCCESS)
-					{
-						oracleCloseStatement(session, NULL);
-						oracleError(FDW_UNABLE_TO_CREATE_EXECUTION,
-							"error executing query: OCIDescriptorAlloc failed to allocate LOB descriptor",
-							oraMessage);
-					}
 					break;
 				case ORA_TYPE_RAW:
-					/* binary types are retrieved in binary form */
+					/* RAW is retrieved in binary form */
 					type = SQLT_BIN;
 					break;
 				default:
@@ -1630,6 +1656,23 @@ oracleExecuteQuery(oracleSession *session, const char *query, const struct oraTa
 					type = SQLT_STR;
 			}
 
+			/* check if it is a LOB column */
+			if (type == SQLT_BLOB || type == SQLT_BFILE || type == SQLT_CLOB)
+			{
+				/* allocate a LOB locator, store a pointer to it in "val" */
+				if (checkerr(
+					OCIDescriptorAlloc((const dvoid *)session->envhp,
+						(dvoid **)oraTable->cols[i]->val,
+						OCI_DTYPE_LOB, 0, NULL),
+					(dvoid *)session->envhp, OCI_HTYPE_ENV) != OCI_SUCCESS)
+				{
+					oracleCloseStatement(session, NULL);
+					oracleError(FDW_UNABLE_TO_CREATE_EXECUTION,
+						"error executing query: OCIDescriptorAlloc failed to allocate LOB descriptor",
+						oraMessage);
+				}
+			}
+
 			/* define result value */
 			defnhp = NULL;
 			if (checkerr(
@@ -1725,7 +1768,7 @@ oracleFetchNext(oracleSession *session, struct oraTable *oraTable)
  * 		"oraTable" is passed so that all LOB locators can be freed in case of errors.
  */
 void
-oracleGetLob(oracleSession *session, struct oraTable *oraTable, void *locptr, char **value, long *value_len)
+oracleGetLob(oracleSession *session, struct oraTable *oraTable, void *locptr, oraType type, char **value, long *value_len)
 {
 	OCILobLocator *locp = *(OCILobLocator **)locptr;
 	ub4 amount;
@@ -1734,6 +1777,22 @@ oracleGetLob(oracleSession *session, struct oraTable *oraTable, void *locptr, ch
 	/* initialize result buffer length */
 	*value_len = 0;
 
+	/* open the LOB if it is a BFILE */
+	if (type == ORA_TYPE_BFILE)
+	{
+		result = checkerr(
+			OCILobFileOpen(session->svchp, session->errhp, locp, OCI_FILE_READONLY),
+			(dvoid *)session->errhp, OCI_HTYPE_ERROR);
+
+		if (result == OCI_ERROR)
+		{
+			oracleCloseStatement(session, oraTable);
+			oracleError(FDW_UNABLE_TO_CREATE_EXECUTION,
+				"error fetching result: OCILobFileOpen failed to open BFILE",
+				oraMessage);
+		}
+	}
+
 	/* read the LOB in chunks */
 	do
 	{
@@ -1769,6 +1828,22 @@ oracleGetLob(oracleSession *session, struct oraTable *oraTable, void *locptr, ch
 
 	/* string end for CLOBs */
 	(*value)[*value_len] = '\0';
+
+	/* close the LOB if it is a BFILE */
+	if (type == ORA_TYPE_BFILE)
+	{
+		result = checkerr(
+			OCILobFileClose(session->svchp, session->errhp, locp),
+			(dvoid *)session->errhp, OCI_HTYPE_ERROR);
+
+		if (result == OCI_ERROR)
+		{
+			oracleCloseStatement(session, oraTable);
+			oracleError(FDW_UNABLE_TO_CREATE_EXECUTION,
+				"error fetching result: OCILobFileClose failed to close BFILE",
+				oraMessage);
+		}
+	}
 }
 
 /*
@@ -1843,6 +1918,7 @@ closeSession(OCIEnv *envhp, OCIServer *srvhp, OCISession *userhp, int disconnect
 	struct envEntry *envp;
 	struct srvEntry *srvp;
 	struct connEntry *connp, *prevconnp = NULL;
+	OCITrans *txnhp = NULL;
 
 	/* search environment handle in cache */
 	for (envp = envlist; envp != NULL; envp = envp->next)
@@ -1904,9 +1980,23 @@ closeSession(OCIEnv *envhp, OCIServer *srvhp, OCISession *userhp, int disconnect
 	/* free the session handle */
 	(void)OCIHandleFree((dvoid *)connp->userhp, OCI_HTYPE_SESSION);
 
+	/* get the transaction handle */
+	if (!silent && checkerr(
+		OCIAttrGet((dvoid *)connp->svchp, (ub4)OCI_HTYPE_SVCCTX,
+			(dvoid *)&txnhp, (ub4 *)0, (ub4)OCI_ATTR_TRANS, envp->errhp),
+		(dvoid *)envp->errhp, OCI_HTYPE_ERROR) != OCI_SUCCESS)
+	{
+		oracleError(FDW_UNABLE_TO_CREATE_REPLY,
+			"error closing session: OCIAttrGet failed to get transaction handle",
+			oraMessage);
+	}
+
 	/* free the service handle */
 	(void)OCIHandleFree((dvoid *)connp->svchp, OCI_HTYPE_SVCCTX);
 
+	/* free the transaction handle */
+	(void)OCIHandleFree((dvoid *)txnhp, OCI_HTYPE_TRANS);
+
 	/* remove the session handle from the cache */
 	if (prevconnp == NULL)
 		srvp->connlist = connp->next;
